import { useReducer } from 'react';

interface State<T> {
  key: string;
  version: string;
  storage: Storage;
  data: T;
}
interface Action<T> {
  type: 'set' | 'remove';
  data?: T;
}

function formatStorageKey(key: string, version: string) {
  return `${key}[${version}]`;
}

function reducer<T>(state: State<T>, action: Action<T>) {
  const updatedAt = new Date();
  switch (action.type) {
    case 'set': {
      const payload = {
        data: action.data,
        updatedAt: updatedAt.toISOString(),
      };
      const json = JSON.stringify(payload);
      state.storage.setItem(formatStorageKey(state.key, state.version), json);
      break;
    }
    case 'remove': {
      state.storage.removeItem(formatStorageKey(state.key, state.version));
      break;
    }
    default:
      break;
  }
  return {
    ...state,
    updatedAt,
    data: action.data || null,
  };
}

function getItemFromStorage(key: string, version: string, storage: Storage) {
  try {
    const json = storage.getItem(formatStorageKey(key, version));
    if (!json) return { data: null, updatedAt: new Date() };
    const payload = JSON.parse(json!);
    const data = payload?.data ?? {};
    const updatedAt = new Date(payload?.version ?? new Date().toISOString());
    return { data, updatedAt };
  } catch (e) {
    console.error(e);
    return { data: null, updatedAt: new Date() };
  }
}

export function useBrowserStorage<T>(
  storageKey: string,
  dataVersion: string,
  storageInterface: Storage,
) {
  const [state, dispatch] = useReducer(
    reducer,
    {
      key: storageKey,
      version: dataVersion,
      storage: storageInterface,
      ...getItemFromStorage(storageKey, dataVersion, storageInterface),
    },
    ({ key, version, storage, updatedAt, data }) => ({
      key,
      version,
      storage,
      updatedAt,
      data,
    }),
  );
  return {
    updatedAt: state.updatedAt,
    key: state.key,
    version: state.version,
    get: () => state.data,
    set: (data: T) => {
      dispatch({ type: 'set', data });
      return data;
    },
    remove: () => {
      dispatch({ type: 'remove' });
    },
  };
}
